#include <SX126x.h>

SX126x LoRa;

// receive data length and pointer
const uint8_t packetLength = 15;
uint8_t packetData[packetLength];
volatile bool flag = false;

void setup() {

  // Begin serial communication
  Serial.begin(38400);

  // Begin LoRa radio and set NSS, reset, busy, txen, and rxen pin with connected arduino pins
  Serial.println("Begin LoRa radio");
  int8_t nssPin = 10, resetPin = 9, busyPin = 4, irqPin = 2, txenPin = 8, rxenPin = 7;
  if (!LoRa.begin(nssPin, resetPin, busyPin, irqPin, txenPin, rxenPin)){
    Serial.println("Something wrong, can't begin LoRa radio");
    while(1);
  }

  // Configure TCXO used in RF module
  Serial.println("Set RF module to use TCXO as clock reference");
  uint8_t dio3Voltage = SX126X_DIO3_OUTPUT_1_8;
  uint32_t tcxoDelay = SX126X_TCXO_DELAY_10;
  LoRa.setDio3TcxoCtrl(dio3Voltage, tcxoDelay);
  
  // Set frequency to 915 Mhz
  Serial.println("Set frequency to 915 Mhz");
  LoRa.setFrequency(915000000);

  // Set RX gain
  Serial.println("Set RX gain to power saving gain");
  LoRa.setRxGain(SX126X_RX_GAIN_POWER_SAVING);

  // Configure modulation parameter including spreading factor (SF), bandwidth (BW), and coding rate (CR)
  Serial.println("Set modulation parameters:\n\tSpreading factor = 7\n\tBandwidth = 125 kHz\n\tCoding rate = 4/5");
  uint8_t sf = 7;
  uint8_t bw = SX126X_LORA_BW_125;
  uint8_t cr = SX126X_LORA_CR_4_5;
  LoRa.setLoRaModulation(sf, bw, cr);

  // Configure packet parameter including header type, preamble length, payload length, and CRC type
  Serial.println("Set packet parameters:\n\tExplicit header type\n\tPreamble length = 12\n\tPayload Length = 15\n\tCRC on");
  uint8_t headerType = SX126X_LORA_HEADER_EXPLICIT;
  uint16_t preambleLength = 12;
  uint8_t payloadLength = 15;
  uint8_t crcType = SX126X_LORA_CRC_ON;
  LoRa.setLoRaPacket(headerType, preambleLength, payloadLength, crcType);

  // Set syncronize word for public network (0x3444)
  Serial.println("Set syncronize word to 0x3444");
  LoRa.setLoRaSyncWord(0x3444);

  Serial.println("\n-- LoRa RECEIVER --\n");

  // Register callback function to be called every RX done
  LoRa.onReceive(getReceiveData);

  // Begin request LoRa packet in continuous mode
  LoRa.request(SX126X_RX_MODE_CONTINUOUS);
}

void loop() {

  if (flag) {
    // Print received package
    Serial.write(packetData, packetLength - 1);
    Serial.print("  ");
    Serial.println(packetData[packetLength - 1]);

    // Print packet/signal status including RSSI, SNR, and signalRSSI
    Serial.print("Packet status: RSSI = ");
    Serial.print(LoRa.rssi());
    Serial.print(" dBm | SNR = ");
    Serial.print(LoRa.snr());
    Serial.println(" dB");
    Serial.println();

    // Reset flag
    flag = false;
  }
}

void getReceiveData() {

  // set flag
  flag = true;
  // Store received data
  for (uint8_t i=0; i<packetLength; i++) {
    packetData[i] = LoRa.read();
  }
}
